@@ -1,14 +1,23 @@
 import Cookie
-import datetime
+from datetime import datetime, timedelta
 import hmac
 import md5
 import os
 import random
 import re
+import sys
 import time
 import UserDict
+import warnings
+
+try:
+    from paste.registry import StackedObjectProxy
+    beaker_session = StackedObjectProxy(name="Beaker Session")
+except:
+    beaker_session = None
 
 from beaker.container import namespace_registry
+from beaker.util import coerce_session_params
 
 __all__ = ['SignedCookie', 'Session']
 
@@ -105,10 +114,10 @@ class Session(UserDict.DictMixin):
             self.cookie[self.key]['path'] = '/'
             if self.cookie_expires is not True:
                 if self.cookie_expires is False:
-                    expires = datetime.datetime.fromtimestamp( 0x7FFFFFFF )
-                elif isinstance(self.cookie_expires, datetime.timedelta):
-                    expires = datetime.datetime.today() + self.cookie_expires
-                elif isinstance(self.cookie_expires, datetime.datetime):
+                    expires = datetime.fromtimestamp( 0x7FFFFFFF )
+                elif isinstance(self.cookie_expires, timedelta):
+                    expires = datetime.today() + self.cookie_expires
+                elif isinstance(self.cookie_expires, datetime):
                     expires = self.cookie_expires
                 else:
                     raise ValueError("Invalid argument for cookie_expires: %s"
@@ -245,3 +254,136 @@ class Session(UserDict.DictMixin):
     def debug(self, message):
         if self.log_file is not None:
             self.log_file.write(message)
+
+class SessionObject(object):
+    """Session proxy/lazy creator
+    
+    This object proxies access to the actual session object, so that in the
+    case that the session hasn't been used before, it will be setup. This
+    avoid creating and loading the session from persistent storage unless
+    its actually used during the request.
+    
+    """
+    def __init__(self, environ, **params):
+        self.__dict__['_params'] = params
+        self.__dict__['_environ'] = environ
+        self.__dict__['_sess'] = None
+        self.__dict__['_headers'] = []
+    
+    def _session(self):
+        """Lazy initial creation of session object"""
+        if self.__dict__['_sess'] is None:
+            params = self.__dict__['_params']
+            environ = self.__dict__['_environ']
+            self.__dict__['_headers'] = req = {'cookie_out':None}
+            req['cookie'] = environ.get('HTTP_COOKIE')
+            self.__dict__['_sess'] = Session(req, use_cookies=True, **params)
+        return self.__dict__['_sess']
+    
+    def __getattr__(self, attr):
+        return getattr(self._session(), attr)
+    
+    def __setattr__(self, attr, value):
+        setattr(self._session(), attr, value)
+    
+    def __delattr__(self, name):
+        self._session().__delattr__(name)
+    
+    def __getitem__(self, key):
+        return self._session()[key]
+    
+    def __setitem__(self, key, value):
+        self._session()[key] = value
+    
+    def __delitem__(self, key):
+        self._session().__delitem__(key)
+    
+    def __repr__(self):
+        return self._session().__repr__()
+    
+    def __iter__(self):
+        """Only works for proxying to a dict"""
+        return iter(self._session().keys())
+    
+    def __contains__(self, key):
+        return self._session().has_key(key)
+
+class SessionMiddleware(object):
+    deprecated = True
+    session = beaker_session
+    
+    def __init__(self, wrap_app, config=None, environ_key='beaker.session', **kwargs):
+        """Initialize the Session Middleware
+        
+        The Session middleware will make a lazy session instance available 
+        every request under the ``environ['beaker.cache']`` key by default. The location in
+        environ can be changed by setting ``environ_key``.
+        
+        ``config``
+            dict  All settings should be prefixed by 'cache.'. This method of
+            passing variables is intended for Paste and other setups that
+            accumulate multiple component settings in a single dictionary. If
+            config contains *no cache. prefixed args*, then *all* of the config
+            options will be used to intialize the Cache objects.
+        
+        ``environ_key``
+            Location where the Session instance will keyed in the WSGI environ
+        
+        ``**kwargs``
+            All keyword arguments are assumed to be cache settings and will
+            override any settings found in ``config``
+        """
+        if self.deprecated:
+            warnings.warn('SessionMiddleware is moving to beaker.middleware in '
+              '0.8', DeprecationWarning, 2)
+        
+        config = config or {}
+        
+        # Load up the default params
+        self.options = dict(invalidate_corrupt=False, type=None, 
+                           data_dir=None, key='beaker.session.id', 
+                           timeout=None, secret=None, log_file=None)
+
+        # Pull out any config args meant for beaker session. if there are any
+        for key, val in config.iteritems():
+            if key.startswith('beaker.session.'):
+                self.options[key[15:]] = val
+            if key.startswith('session.'):
+                self.options[key[8:]] = val
+        
+        # Coerce and validate session params
+        coerce_session_params(config)
+        
+        # Update the params with the ones passed in
+        self.options.update(config)
+        self.options.update(kwargs)
+        
+        self.wrap_app = wrap_app
+        self.environ_key = environ_key
+        
+    def __call__(self, environ, start_response):
+        session = SessionObject(environ, **self.options)
+        if environ.get('paste.registry'):
+            environ['paste.registry'].register(self.session, session)
+        environ[self.environ_key] = session
+        
+        def session_start_response(status, headers, exc_info = None):
+            if session.__dict__['_sess'] is not None:
+                cookie = session.__dict__['_headers']['cookie_out']
+                if cookie:
+                    headers.append(('Set-cookie', cookie))
+            return start_response(status, headers, exc_info)
+        try:
+            response = self.wrap_app(environ, session_start_response)
+        except:
+            ty, val = sys.exc_info()[:2]
+            if isinstance(ty, str):
+                raise ty, val, sys.exc_info()[2]
+            if ty.__name__ == 'HTTPFound' and \
+                    session.__dict__['_sess'] is not None:
+                cookie = session.__dict__['_headers']['cookie_out']
+                if cookie:
+                    val.headers.append(('Set-cookie', cookie))
+            raise ty, val, sys.exc_info()[2]
+        else:
+            return response
